#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
.. codeauthor:: Nathan Hertz <nhertz@nrao.edu>

File containing definitions for one half of the capability side of the Workspaces REST API,
concerning capabilities themselves
"""

import http

from pycapo import CapoConfig

from pyramid.httpexceptions import (
    HTTPBadRequest,
    HTTPExpectationFailed,
    HTTPNotFound,
    HTTPPreconditionFailed,
)
from pyramid.request import Request
from pyramid.response import Response
from pyramid.view import view_config

from workspaces.capability.schema import Capability


@view_config(route_name="healthcheck", renderer="json")
def get_healthcheck(request: Request) -> Response:
    """
    Route for capability container to check notification is ready
    :param request: GET request
    :return:
    """
    return Response(
        status=http.HTTPStatus.OK, json_body={"healthcheck": f"Capability service returned {http.HTTPStatus.OK}"}
    )


@view_config(route_name="view_capability", renderer="json")
def view_capability(request: Request) -> Response:
    """
    Pyramid view that accepts a request to view a capability and responds with the requested capability's
    info, if it exists
    URL: capability/{capability_name}

    :param request: GET request
    :return: Response with JSON-formatted capability info if found
        or 404 response (HTTPNotFound)
    """
    capability = request.capability_info.lookup_capability(request.matchdict["capability_name"])
    if capability:
        return Response(json_body=capability.__json__())
    else:
        return HTTPNotFound(detail=f"Capability {request.matchdict['capability_name']} not found.")


@view_config(route_name="view_enabled_capabilities", renderer="json")
def view_enabled_capabilities(request: Request) -> Response:
    """
    Gets the names of Capabilities that are enabled
    URL: /capabilities

    :param request: GET request
    :return: List of the names of enabled Capabilities
        or 404 response (HTTPNotFound)
    """
    capabilities = request.capability_info.get_enabled_capabilities()
    if capabilities:
        return Response(json_body=capabilities)
    else:
        return HTTPNotFound(detail=f"No enabled capabilities found.")


# TODO: Fix out of date endpoint. Make it use the create_capability method in info
@view_config(route_name="create_capability", renderer="json")
def create_capability(request: Request) -> Response:
    """
    Pyramid view that accepts a request to create a capability
    URL: capability/create

    :param request: POST request, expecting JSON parameters ["capability_name", "max_jobs"]
        optionally, accepts a boolean "enabled" parameter
    :return: Response with JSON-formatted capability info of newly created capability
        or 400 response (HTTPBadRequest) if expected parameters not given
        or 412 response (HTTPPreconditionFailed) if capability with given name already exists
    """
    expected_params = ["capability_name", "max_jobs", "has_image_products"]
    params = request.json_body

    if not all([expected in params for expected in expected_params]):
        # JSON params do not contain all expected params
        params_not_given_msg = f"Expected JSON parameters {expected_params}. Received only {params}."
        return HTTPBadRequest(detail=params_not_given_msg)
    elif request.capability_info.lookup_capability(params["capability_name"]):
        # Capability with given name already exists
        already_exists_msg = (
            f"Capability {params['capability_name']} already exists.",
            f"To instead edit existing capability, use capability/{params['capability_name']}/edit.",
        )
        return HTTPPreconditionFailed(detail=already_exists_msg)
    else:
        new_capability = Capability(
            name=params["capability_name"],
            max_jobs=params["max_jobs"],
            enabled=params.get("enabled", True),
            has_image_products=params.get("has_image_products", False),
            paused=params.get("paused", False),
        )
        request.capability_info.save_entity(new_capability)
        return Response(json_body=new_capability.__json__())


@view_config(route_name="edit_capability", renderer="json")
def edit_capability(request: Request) -> Response:
    """
    Pyramid view that accepts a request to edit a capability
    URL: capability/{capability_name}/edit

    :param request: POST request, expecting JSON parameters ["capability_name", "max_jobs"]
    :return: Response with JSON-formatted capability info of newly created capability
        or 400 response (HTTPBadRequest) if expected parameters not given
        or 412 response (HTTPPreconditionFailed) if capability with given name does not exist
        or 417 response (HTTPExpectationFailed) if the capability was unable to be edited

    TODO: In the future, we should check if there are any active requests for the capability being edited and
        disallow its editing until they are finished
    """
    expected_params = ["max_jobs"]
    capability_name = request.matchdict["capability_name"]
    params = request.json_body

    if not any([expected in params for expected in expected_params]):
        # JSON params do not contain any expected params
        params_not_given_msg = f"Expected one or more JSON parameters {expected_params}. Received only {params}."
        return HTTPBadRequest(detail=params_not_given_msg)
    elif not request.capability_info.lookup_capability(capability_name):
        # Capability with given name does not exist
        does_not_exist_msg = (
            f"Capability {capability_name} does not exist.",
            f"To instead create a new capability, use capability/create with the same parameters.",
        )
        return HTTPPreconditionFailed(detail=does_not_exist_msg)
    else:
        max_jobs = params.get("max_jobs", None)
        success = request.capability_info.edit_capability(capability_name, max_jobs)
        if success:
            return Response(body="Capability successfully edited!")
        else:
            return HTTPExpectationFailed(detail=f"Unable to edit capability {capability_name}.")


@view_config(route_name="enable_capability", renderer="json")
def enable_capability(request: Request) -> Response:
    """
    Pyramid view that enables a capability

    :param request: POST request
    :return: HTTP 200 response
        or 412 response (HTTPPreconditionFailed) if capability with given name does not exist
        or 417 response (HTTPExpectationFailed) if the capability was unable to be enabled
    """
    capability_name = request.matchdict["capability_name"]

    if not request.capability_info.lookup_capability(capability_name):
        # Capability with given name does not exist
        does_not_exist_msg = f"Capability {capability_name} does not exist. Cannot enable."
        return HTTPPreconditionFailed(detail=does_not_exist_msg)
    else:
        success = request.capability_info.edit_capability(capability_name, enabled=True)
        if success:
            return Response(body="Capability successfully enabled!")
        else:
            return HTTPExpectationFailed(detail=f"Unable to enable capability {capability_name}.")


@view_config(route_name="disable_capability", renderer="json")
def disable_capability(request: Request) -> Response:
    """
    Pyramid view that disables a capability

    :param request: POST request
    :return: HTTP 200 response
        or 412 response (HTTPPreconditionFailed) if capability with given name does not exist
        or 417 response (HTTPExpectationFailed) if the capability was unable to be disabled
    """
    capability_name = request.matchdict["capability_name"]

    if not request.capability_info.lookup_capability(capability_name):
        # Capability with given name does not exist
        does_not_exist_msg = f"Capability {capability_name} does not exist. Cannot disable."
        return HTTPPreconditionFailed(detail=does_not_exist_msg)
    else:
        success = request.capability_info.edit_capability(capability_name, enabled=False)
        if success:
            return Response(body="Capability successfully disabled!")
        else:
            return HTTPExpectationFailed(detail=f"Unable to disable capability {capability_name}.")


@view_config(route_name="pause_capability", renderer="json")
def pause_capability(request: Request) -> Response:
    """
    Pyramid view that pauses a capability

    :param request: POST request
    :return: HTTP 200 response
        or 412 response (HTTPPreconditionFailed) if capability with given name does not exist
    """
    capability_name = request.matchdict["capability_name"]

    if not request.capability_info.lookup_capability(capability_name):
        # Capability with given name does not exist
        does_not_exist_msg = f"Capability {capability_name} does not exist. Cannot enable."
        return HTTPPreconditionFailed(detail=does_not_exist_msg)
    else:
        request.capability_info.pause_capability(capability_name)
        return Response(status_code=http.HTTPStatus.OK)


@view_config(route_name="unpause_capability", renderer="json")
def unpause_capability(request: Request) -> Response:
    """
    Pyramid view that unpauses a capability

    :param request: POST request
    :return: HTTP 200 response
        or 412 response (HTTPPreconditionFailed) if capability with given name does not exist
    """
    capability_name = request.matchdict["capability_name"]

    if not request.capability_info.lookup_capability(capability_name):
        # Capability with given name does not exist
        does_not_exist_msg = f"Capability {capability_name} does not exist. Cannot enable."
        return HTTPPreconditionFailed(detail=does_not_exist_msg)
    else:
        request.capability_info.unpause_capability(capability_name)
        request.capability_queue_restarter.restart_capability_queue()
        return Response(status_code=http.HTTPStatus.OK)


@view_config(route_name="retrieve_available_qa_staff_in_group", renderer="json")
def retrieve_available_qa_staff_in_group(request: Request) -> Response:
    """
    Gets a list of staff members, by group, available to be assigned as QA
    URL: /capabilities/available_qa_staff?group=

    :param request: GET request
    :return: List of available staff members
        or 404 response (HTTPNotFound)
    """
    available_staff = request.capability_info.get_available_qa_staff_in_group(request.params["group"])
    if available_staff:
        return Response(json_body={f"{request.params['group']}": available_staff})
    else:
        return HTTPNotFound(detail=f"No available staff found.")


@view_config(route_name="get_analyst_email", renderer="json")
def get_analyst_email(request: Request) -> Response:
    """
    Pyramid view that gets the workspaces-analyst email address from the capo config.

    :param request: GET request
    :return: Response containing the workspaces-analyst email address from the capo config
        or 404 response (HTTPNotFound) if it isn't found in the capo config
    """
    analyst_email = CapoConfig().settings("edu.nrao.workspaces.NotificationSettings").analystEmail
    if analyst_email:
        return Response(status_int=http.HTTPStatus.OK, json_body={"resp": f"{analyst_email}"})
    else:
        return HTTPNotFound(detail=f"No workspaces-analyst email found.")
